% This code is licensed under the 3-clause BSD license.
% Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.
% See LICENSE.txt for details.

%%% Uncertainty Propagation %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if B > 1
  fprintf('Solving ensemble of kinetic models...\n\n')
else
  fprintf('Solving kinetic model...\n\n')
end

for b = 1:B

  fprintf('Integration %d of %d\r',b,B); if b == B, fprintf('\n'); end

  ys(:,:,b)     = solver(rxns,k(:,b),y0,tspan,ts,tol);
  Y             = [Y;ys(end,:,b) * mass];
  [g_inf g_net ...
   f_pos f_neg] = rates(rxns,k(:,b),ys(:,:,b),ts);
  g_max(:,b)    = max(g_inf,[],2) .* ~spec(:,4);
  G(:,b)        = 0.5 * (g_net(:,1:end-1) + g_net(:,2:end))  * dt ;
  F_pos(:,:,b)  = 0.5 * (f_pos(:,1:end-1) + f_pos(:,2:end)) .* dt';
  F_neg(:,:,b)  = 0.5 * (f_neg(:,1:end-1) + f_neg(:,2:end)) .* dt';

end

deltaB = metricB_u(ys,U);
DeltaB = metricB(ys,dt,t_max);

[g_max_sort,g_max_ranking] = sort(rmse(g_max,2),'descend');

%%% Network Reduction %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if f_analys

  fprintf('Coarse flux analysis (CFA) starts...\n\n')

  G       = G * N;
  G_quant = quantile(G,(1+clevel)/2,2);

  n_idx_cfa = unique([inits(:,1);spec(G_quant > G_min)]);
  l_idx_cfa = [];

  for l = 1:L
    if prod(sum([0;n_idx_cfa] == rxns(l,:)))
      l_idx_cfa = [l_idx_cfa;l];
    end
  end

  if isempty(l_idx_cfa)
    N_cfa     = N;
    L_cfa     = L;
    ys_cfa    = ys;
    delta_cfa = zeros(B,1);
    Delta_cfa = zeros(B,1);
    l_idx_cfa = (1:L)';
  else
    N_cfa = length(unique(rxns(l_idx_cfa,:))) ...
          - sum(~unique(rxns(l_idx_cfa,:)));
    L_cfa = size(rxns(l_idx_cfa,:),1);
    for b = 1:B
      fprintf('Integration %d of %d\r',b,B); if b == B, fprintf('\n'); end
      ys_cfa(:,:,b)  = solver(rxns(l_idx_cfa,:),[k(l_idx_cfa,b);...
                              k(L+l_idx_cfa,b)],y0,tspan,ts,tol);
      Y_cfa(b,1)     = ys_cfa(end,:,b) * mass;
      Y              = [Y;ys_cfa(end,:,b) * mass];
      delta_cfa(b,1) = metric_u(ys(:,:,b),ys_cfa(:,:,b),U);
      Delta_cfa(b,1) = metric(ys(:,:,b),ys_cfa(:,:,b),dt,t_max);
    end
  end 

  dDelta_cfa       = max(delta_cfa,Delta_cfa);
  dDelta_cfa_quant = quantile(dDelta_cfa,(1+clevel)/2); 

  if dDelta_cfa_quant > y_min

    fprintf('Warning: CFA-based network reduction not reliable!\n\n')
    fprintf('Detailed flux analysis (DFA) starts:\n')
    fprintf('- preparatory step...\n\n')

    for l = 1:L
      k_dfa = [k(1:l-1,1:B);zeros(1,B);k(l+1:L+l-1,1:B);...
               zeros(1,B);k(L+l+1:end,1:B)];
      for b = 1:B
        fprintf('Integration %d of %d\r',B*(l-1)+b,B*L); if b == B && ...
                                                            l == L,...
        fprintf('\n'); end
        ys_dfa_tmp          = solver(rxns,k_dfa(:,b),y0,tspan,ts,tol);
        delta_dfa_prep(l,b) = metric_u(ys(:,:,b),ys_dfa_tmp,U);
        Delta_dfa_prep(l,b) = metric(ys(:,:,b),ys_dfa_tmp,dt,t_max);
      end
    end

    dDelta_dfa_prep       = max(delta_dfa_prep,Delta_dfa_prep);
    dDelta_dfa_prep_quant = quantile(dDelta_dfa_prep,(1+clevel)/2,2);

    fprintf('- control step...\n\n')

    l_idx_dfa = find(dDelta_dfa_prep_quant > y_min / N);

    if isempty(l_idx_dfa)
      N_dfa     = N;
      L_dfa     = L;
      ys_dfa    = ys;
      delta_dfa = zeros(B,1);
      Delta_dfa = zeros(B,1);
      l_idx_dfa = (1:L)';
    else
      N_dfa = length(unique(rxns(l_idx_dfa,:))) ...
            - sum(~unique(rxns(l_idx_dfa,:)));
      L_dfa = size(rxns(l_idx_dfa,:),1);
      for b = 1:B
        fprintf('Integration %d of %d\r',b,B); if b == B, fprintf('\n'); end
        ys_dfa(:,:,b)  = solver(rxns(l_idx_dfa,:),[k(l_idx_dfa,b);...
                                k(L+l_idx_dfa,b)],y0,tspan,ts,tol);
        Y_dfa(b,1)     = ys_dfa(end,:,b) * mass;
        Y              = [Y;ys_dfa(end,:,b) * mass];
        delta_dfa(b,1) = metric_u(ys(:,:,b),ys_dfa(:,:,b),U);
        Delta_dfa(b,1) = metric(ys(:,:,b),ys_dfa(:,:,b),dt,t_max);
      end
    end

    dDelta_dfa       = max(delta_dfa,Delta_dfa);
    dDelta_dfa_quant = quantile(dDelta_dfa,(1+clevel)/2);

    if dDelta_dfa_quant > y_min

      fprintf('Warning: DFA-based network reduction not reliable!\n')
      fprintf('KiNetX cannot suggest a reduced network!\n\n')

      N_        = N;
      L_        = L;
      l_idx_xfa = (1:L)';

    else

      N_        = N_dfa;
      L_        = L_dfa;
      l_idx_xfa = l_idx_dfa;

      fprintf('Good news: DFA-based network reduction is reliable!\n\n')
      fprintf('Number of nodes has been reduced from %d to %d.\n',N,N_)    
      fprintf('Number of edges has been reduced from %d to %d.\n\n',L,L_)

    end

  else

    N_        = N_cfa;
    L_        = L_cfa;
    l_idx_xfa = l_idx_cfa;

    fprintf('Good news: CFA-based network reduction is reliable!\n')
    fprintf('Number of nodes has been reduced from %d to %d.\n',N,N_)    
    fprintf('Number of edges has been reduced from %d to %d.\n\n',L,L_)    

  end

  rxns_ = rxns(l_idx_xfa,:);
  k_    = [k(l_idx_xfa,:);k(L+l_idx_xfa,:)];  
  spec_ = nodes(2:max(rxns_(:))+1,:);
  
end

%%% Sensitivity Analysis %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if s_analys

  if ~f_analys
    N_    = N;
    L_    = L;
    rxns_ = rxns;
    k_    = k;
    spec_ = spec;
  end

  fprintf('Morris-type sensitivity analysis...\n\n')

  i_msa = 0;

  for c = 1:C 

    k_current = k_(:,c);

    for l = 1:L_

      i_msa = i_msa + 1;

      fprintf('Integration %d of %d\r',i_msa,C*L_); if i_msa == C*L_,...
      fprintf('\n'); end

      k_current(l)    = k_(l,c+1);
      k_current(L_+l) = k_(L_+l,c+1);

      ys(:,:,c,l+1)  = solver(rxns_,k_current,y0,tspan,ts,tol);
      Y_msa(l,c)     = ys(end,:,c,l+1) * mass;
      Y              = [Y;ys(end,:,c,l+1) * mass];
      delta_msa(l,c) = metric_u(ys(:,:,c,l),ys(:,:,c,l+1),U);
      Delta_msa(l,c) = metric(ys(:,:,c,l),ys(:,:,c,l+1),dt,t_max);

    end

  end

  Y_msa = Y_msa(:);

  dDelta_msa       = max(delta_msa,Delta_msa);
  dDelta_msa_0     = dDelta_msa(:,1); 
  dDelta_msa_quant = quantile(dDelta_msa,(1+clevel)/2,2);

  l_idx_msa_0 = find(dDelta_msa_0 > y_min);
  l_idx_msa   = find(dDelta_msa_quant > y_min);

end
